#!/bin/sh
#set -x
#
# CrossControl - Device updater
#
# Update existing CrossCore devices
# 
# Below V is updater identification, increment with update (might become usefull someday)
V=AI
#
# Version is not standard a.b.c -number since we might have multiple branches and
# I dont want script to be directly assosiated to any CrossCore release. 

# 230418 TK,  AH: Added U-boot update
# 270917 TK,  AG: Converted for VI2.
# 130617 TK,  AF: Added rescue system support for VS
# 300317 TK,  AE: Converted for VS.
# 140512 TK,  AD: Converted for XM2. XA support removed from this version
# 130115 TK,  AC: Cross XA - Added option to disable reboot after update
# 120514 HB,  AB:Cross XA - Add SD support, NAND support dropped, also drop support for XA1+CCP XA proto
# 120206 HB,  AB:Cross XA - Support for i.MX53 SABRE-AI
# 110815 TKe, AA:Cross XA - Fixed mtd devices for CCP XA proto. Added U-Boot
# 100623 TTH, A9:Cross XA  - Multi environment, with CCP XA proto
# 100518 TTH, A8.2:Cross XA1 - CC+under-temp-fix 
# 100322 TTH, A8.1:Cross XA1 - Bootstrap flash oops
# 100318 TTH, A8:Cross XA1 - With FPGA
# 1003xx TTH, A7:Cross XA1 - Cleaned and structured, prepared for FPGA
# 100224 TTH, A6:Cross XA1 - RW-mode protection
# 100127 TTH, A5:Cross XA1 - Using backup-system term
# 091120 TTH, A4:Cross XA1 - reboot-tune-up
# 091117 TTH, A3:Cross XA1 - With bootloader
# 091106 TTH, A2:Cross XA1 - Secondary System support
# 091026 TTH, A1:Cross XA1 - Proto environment version
#    -- branch point for new device --
# 090422 MNK, Now with ramimage system support as well
# 090211 TTH, Now with system image and separate files and CF image
# 090202 TTH, Now with bootloader
# 0812.. TTH, 
#
#
#

# Help usage info
show_help()
{
if [ ! "x$1" == "x" ] ; then 
    echo -e "\nErr:"$1"\n" ; 
    echo -e " Run with -h for help and details.\n"
    exit 4
fi
cat <<EOF
 Reflash CrossControl device with new firmware, distribution and addons.

Usage: `basename $0` <-p|-s> [-b file] [-k file] [-r file] [[-c file] [-d]]
Usage: `basename $0` -l mcs-file

  Choice of update target:
   -p   Update Primary side images (default)
   -s   Update Backup-System images
   -l   Update FPGA Logic code (not together with others)
   -c   Addons package to extract to CF (add -p or -s if want both) 

  Default names used unless overwritten with command line:
   -b   Bootloader:  ccpilot_vi2_u-boot.bin
   -k   Kernel:      ccpilot-vi2_kernel.bin
   -r   Rootfs:      ccpilot-vi2_rootfs.bin

   -d   When extracting CF-pack, clear included folders
   -f   Force-flag disables all user interaction, use for scripting.
   -m   Disable MD5 sum matching - DANGEROUS
   -u   Clear /opt/*
   -2   Do reboot-rescue instead of reboot after flashing 
   -3   Do not reboot after flashing

   -v   Verbose messages
   -q   Quiet messages

 NOTE: Actions of this tool can be destructive. Use with care.
       Shutdown all excess applications BEFORE executing this update.
       Stop all file-system access and prevent preliminary power loss.

EOF
 
 exit 4
}

## Return member from array (busybox shell does not handle real array variables)
## 1 being the first item.
# args: <id>, <array>
get( ) {
 RV= 
 if [ $1 -eq 1 ] ; then RV=$2; fi
 if [ $1 -eq 2 ] ; then RV=$3; fi
 if [ $1 -eq 3 ] ; then RV=$4; fi
 if [ $1 -eq 4 ] ; then RV=$5; fi
 if [ $1 -eq 5 ] ; then RV=$6; fi
 if [ ${#RV} -eq 0 ] ; then echo "Internal error with array $*"; exit 1 ; fi
 return ${#RV}
}

## Conditional Leveled echos, level 1=always, 2=normally, 3=in verbose
# args: <level> <what to echo>
decho (){
 EL=$1
 shift
 if [ $EL -lt $DL ] ; then
       /bin/echo "$*"
 fi
}
dechon (){
 EL=$1
 shift
 if [ $EL -lt $DL ] ; then
     /bin/echo -n "$*"
 fi
}
dechoe (){
 EL=$1
 shift
 if [ $EL -lt $DL ] ; then
     /bin/echo -e "$*"
 fi
}
dechoen (){
 EL=$1
 shift
 if [ $EL -lt $DL ] ; then
     /bin/echo -en "$*"
 fi
}

# Prepare things for FPGA update
# TBD - convert this to use ccsettingsconsole
prepare_fpga()
{
  ferr=0
  decho 1 "ERROR: FPGA update not supported. - Quiting.."
  exit 3

  if [ $ferr -eq 1 ] ; then
    decho 1 "ERROR: Something wrong while preparing for FPGA update. - Quiting.."
    exit 3
  fi

}

# Do the FPGA
flash_fpga()
{
    decho 1 "ERROR: FPGA update not supported. - Quiting.."
    exit 3

    echo "Erasing and Flashing FPGA" && \
    ./mcstool -f $TMPFBIN && \
    echo "Verifying the flashed data" && \
    ./mcstool -r $TMPFBIN.out `ls -l $TMPFBIN | tr -s \  | cut -d\  -f5` >/dev/null && \
    ./mcstool -m $TMPFBIN.out readback.mcs >/dev/null && \
    dos2unix readback.mcs && \
    dos2unix $FIMG && \
    diff $FIMG readback.mcs -qs 
    if [ $? -ne 0 ] ; then
        echo -e   "Error: Something went wrong, during the flash !"
        echo -e "\n YOUR FPGA FLASH IS NOW CORRUPT!  FIX IT BEFORE CUTTING POWER!\n"
        echo -e   " If you power off, the device might not reboot. !\n "
        exit 1
    fi
    echo -e "\nResult: FPGA code flashed and verified successfully."
    echo "        The new code is used after next full power off! (not reboot)"
    exit 0  ## Exit with success here, no need to reboot
}

checkFile()
{
 FN=$1
 get $USING $MD5
 MD=$RV
 if [ ! -f "$MD" ] ; then
  if [ $md5 -eq 0 ] ; then 
   decho 3 " WARNING: MD5 file: Not found but Ignoring with \"-m\"" 
   return 0
  else 
   decho 1 "   ERROR: MD5 file: $MD: Not found. - Quiting.."
   exit 5
  fi
 fi
 dechon 2  "      Checking MD5: $FN: "
 
 (grep -E " $FN$" $MD || echo "" ) | md5sum -c - > /dev/null

 if [ $? -ne 0 ] ; then 
  # error
  if [ $md5 -eq 0 ] ; then 
   decho 2 "FAILED! - Ignoring with \"-m\"" 
  else 
   decho 2 "FAILED! - Quiting.." 
   exit 5;
  fi 
 else
  decho 2 "OK"
 fi

}

get_cpu() {
 # Get our place
 id=1
 while  [ $id -le ${SUPCNT} ]  ; do
  get $id $SUPPORTCPU
  KEY=${RV}  
  grep -q $KEY /proc/cpuinfo
  if [ $? -eq 0 ] ; then
     get $id $SUPPORTED
     DEVICE=$RV
     USING=$id
     return 0
  fi
  id=$((id+1))
 done
 return 1
}

check_environment() {

 get_cpu
 if [ $? -ne 0 ] ; then
 show_help "This works only on supported devices: $SUPPORTED"; fi

 echo "   System Hardware: $DEVICE" 

 if [ `id -u` -ne 0 ] ; then show_help "This works ONLY when used as root -user!" ; fi

 this_side=0 # unknown
 # Detect side (normal/backup) based on which rootfs partition was given in command line
 if mount | grep -q mmcblk0p2
 then
  this_side=1 # running from normal rootfs
 elif mount | grep -q mmcblk0p3
 then
  this_side=2 # backup
 else
  echo "Unknown run environment... - Quiting.."
  exit 9
 fi

 # Where are we
 #PWD=`pwd`
 #if [ ${PWD:0:10} != "/usr/local" -o ${PWD:0:4} != "/opt" ] ; then echo "WARNING: Executing outside of /usr/local or /opt, expecting problems..." ; fi
}

set_defaults() {

# Supported Environments idents 
SUPPORTED="ccpilot-vi2"
SUPPORTCPU="ARMv7"
SUPCNT=1
DEVICE=
USING=
# Default image names
MD5="ccpilot-vi2.md5sum"
BIMG=
BIMG1="ccpilot-vi2_u-boot.bin"
KIMG=
RIMG=
FIMG=
KIMG1="ccpilot-vi2_kernel.bin"
RIMG1="ccpilot-vi2_rootfs.bin"
KIMG2=ccpilot-vi2_rescue_kernel.bin
RIMG2=ccpilot-vi2_rescue_rootfs.bin
AIMG=ccpilot-vi2_full.bin
CIMG=
SIMG=

TMPFBIN=/tmp/fpga.bin

# Device names
BDEV=mmcblk0
KDEV=zImage # Filename
RDEV=mmcblk0p2
FDEV=fpga

KDEV1=zImage
RDEV1=mmcblk0p2

# Also backup partition are on same device on SD
KDEV2=zImage_rescue # filename
RDEV2=mmcblk0p3

bi=1
ki=1
ri=1
ai=0
fi=0

si=0
ni=0

ci=0
de=0

usrclear=0
noreboot=0

ask=1
md5=1
update_side=1  # default to main system
update_this=1  # allow it quietly always (?)

DL=3 # Debug/verbose level

}

# Return 1 if argument exists
# Return 0 if argument not exist
get_argument() {
 A=$1
 RV=0
 shift
 CB=`getopt -uq  -- $A "$@"`
 CB=`echo $CB| sed "s/--.*//"`
 if [ ! -z "$CB" ] ; then RV=1 ; fi
}

# Return value of argument
# Return 0 if no/empty
get_argvalue() {
 A=$1
 shift
 #echo ";$@;"
 CB=`getopt -uq -- $A: "$@" | sed "s/^ *[^ ]* \(.*\) -.*/\1/" | sed "s/\-.*//"`
 if [ -z $CB ] ; then  
   RV=0
 fi
 RV=$CB
}

# Return filename argument value and check for existence.
# Quit on not available
get_filearg() {
 A=$1
 shift
 #echo ";$@;"
 CB=`getopt -uq -- $A: "$@" | sed "s/^ *[^ ]* \(.*\) -.*/\1/" | sed "s/\-\-.*//"`
 if [ -z $CB ] ; then  
   show_help "Option -$A requires a file name argument."
 fi
 if [ ! -f "$CB" ] ; then
   show_help "File for -$A ($CB) not found!"
 fi
 RV=$CB
}

# Return filename argument valu and check for existence.
# If argument not avail, ignore.
get_optfilearg() {
 get_argument $@
 if [ $RV -eq 1 ] ; then get_filearg $@ ; else RV="" ; fi
}

# Read all arguments and set variables accordingly
read_arguments() { 

 if [ $# -gt 0 ]
 then

  get_argument q $@
  if [ $RV -eq 1 ] ; then DL=2; fi
  get_argument v $@
  if [ $RV -eq 1 ] ; then DL=4; fi

  # H elp
  get_argument h $@
  if [ $RV -eq 1 ] ; then show_help; fi

  # F orce
  get_argument f $@
  if [ $RV -eq 1 ] ; then ask=0; fi

  # M d5 check
  get_argument m $@
  if [ $RV -eq 1 ] ; then md5=0; fi

  get_argument d $@
  if [ $RV -eq 1 ] ; then de=1; fi

  # U sr/local delete check
  get_argument u $@
  if [ $RV -eq 1 ] ; then 
    usrclear=1; 
    de=0 # no more need for this
  fi

  # Do not reboot after update
  get_argument 3 $@
  if [ $RV -eq 1 ] ; then noreboot=1;  fi

  # Make next reboot -> reboot-rescue
  get_argument 2 $@
  if [ $RV -eq 1 ] ; then 
    echo "1 0" >/dev/factory_reset
  fi

  # Make next reboot-rescue -> reboot
  get_argument 1 $@
  if [ $RV -eq 1 ] ; then 
    #echo "0 0" >/dev/factory_reset
      echo "No factory reset"
  fi

  get_optfilearg c $@
  if [ ! -z "$RV" ] ; then 
    ci=1
    bi=0
    ki=0
    ri=0
    ai=0
    fi=0
    if [ ${RV:0:1} == "/" ] ; then
      CIMG="$RV"
    else
      CIMG="/tmp/$RV"
    fi

    update_side=0;  # remove default 
  fi

  get_argument s $@
  if [ $RV -eq 1 ] ; then 
    update_side=2; 
    bi=1
    ki=1
    ri=1
    ai=0
    fi=0
  fi

  get_argument p $@
  if [ $RV -eq 1 ] ; then 
    update_side=1;
    bi=1
    ki=1
    ri=1
    ai=0
    fi=0
  fi

  get_argument t $@
  if [ $RV -eq 1 ] ; then update_this=1; fi

  # Check for overrided filenames
  get_optfilearg b $@
  if [ ! -z "$RV" ] ; then BIMG="$RV" ; fi
  get_optfilearg k $@
  if [ ! -z "$RV" ] ; then KIMG="$RV" ; fi
  get_optfilearg r $@
  if [ ! -z "$RV" ] ; then RIMG="$RV" ; fi
  get_optfilearg a $@
  if [ ! -z "$RV" ] ; then AIMG="$RV" ; fi

  get_optfilearg l $@
  if [ ! -z "$RV" ] ; then
    FIMG="$RV" 
    update_side=4
    ki=0
    ri=0
    ai=0
    fi=1
  fi
fi # params exist
} #readarguments

# Display info if..
show_bootloader() {
 if [ $bi -eq 1 ] ; then
  if [ -f "$BIMG" ] ; then
   decho 1 "  Requested action: Bootloader update"
   decho 3 "   Bootloader-image: $BIMG"   
  else
   # no Bootloader file, think its ok.
      echo "No bootloader"
   bi=0
  fi
 fi
}

# Display preliminary info of what we are about to do
show_request() {
# sides ok?
 if [ $this_side -eq 1 ] ; then
  decho 1 " Now running on   : normal side"
 else
  decho 1 " Now running on   : BACKUP system"
 fi

 if [ $ci -eq 1 ] ; then
  decho 1  "  Requested action: Addon install"
  decho 3  "         From file: $CIMG"
 fi
 
 if [ $update_side -eq 4 ] ; then
  decho 1  "  Requested action: FPGA update"
  decho 2  "   Used file image: $FIMG"
 elif [ $update_side -eq 2 ] ; then
  if [ -z $KIMG ] ; then KIMG=$KIMG2; fi
  if [ -z $RIMG ] ; then RIMG=$RIMG2; fi
  KDEV=$KDEV2
  RDEV=$RDEV2
  KSEEK=$KSEEK2
  RSEEK=$RSEEK2

  show_bootloader

  dechoe 1 "  Requested action: BACKUP system update"
  dechoe 3 "      kernel-image: $KIMG"
  dechoe 3 "        root-image: $RIMG"

  if [ $this_side -ne 1 ] ; then
   if [ $update_this -eq 1 ] ; then
     decho 3 "Active side update: BY FORCE"
   else
     echo "Active side update: Denyed.";
     exit 7
   fi
  fi


 elif [ $update_side -eq 1 ] ; then
     if [ -z $BIMG ] ; then 
         get $USING $BIMG1 
         BIMG=$RV
     fi
     if [ -z $KIMG ] ; then 
	 get $USING $KIMG1 
	 KIMG=$RV
     fi
  if [ -z $RIMG ] ; then 
    get $USING $RIMG1
    RIMG=$RV
  fi
  get $USING $KDEV1
  KDEV=$RV
  get $USING $RDEV1
  RDEV=$RV

  show_bootloader

  dechoe 1 "  Requested action: NORMAL system update"
  dechoe 3 "      kernel-image: $KIMG"
  dechoe 3 "        root-image: $RIMG"

  if [ $this_side -ne 2 ] ; then
   if [ $update_this -eq 1 ] ; then
     decho 3 "Active side update: BY FORCE"
   else
     echo "Active side update: Denied."
     exit 7
   fi
  fi
   
#  else
#  show_help "Update side not selected. Exit."
#  exit 8
fi 
}

# Acquire sizes of partitions.
get_sizes() {
# No partition available, assume fixed max size of 12MB
SIZEPB=12582912

# No partition available, assume fixed max size of 12MB
SIZEPK=12582912

# Partition might not be available, use fixed size of 251MB
SIZEPRBS=263192576

# Partition might not be available, use fixed size of 743MB
SIZEPR=779059201
}

check_addon_image() {
 if [ ! -f "$CIMG" ] ; then
  echo "ERROR: Bad image? : $CIMG: Not found, skipping.."
  ci=0
 else 
  checkFile $CIMG
  dechon 2 "    Checking addon: $CIMG: "
  tar -tf $CIMG > /tmp/cflist
  if [ $? -ne 0 ] ; then
    decho 1 "Addon is corrupt. Quiting.."
    exit 2
  else 
    DIRS=`cat /tmp/cflist | grep "^[^\/]*\/$" | tr -d \/ `
    decho 2  "OK"
    decho 3  "    addon includes: " `echo $DIRS ` 
  fi
 fi 
}

# CF Addons package install  // 100304, not verified.
addon_install() {

 /bin/echo "Extracting addon package.."
 if [ $de -eq 1 ] ; then
   /bin/echo ".. after clearing included folders."
 fi

 cd /usr/local || exit 9

 if [ $de -eq 1 ] ; then
   rm -rf $DIRS
 fi

 tar -xf $CIMG

 echo -e "\n Addons image installed\n"
 cd /tmp || exit 8
}

# Verify file size range
# args: <Filename> <minsize> <maxsize>
check_file_size() {
 FN=$1
 MIN=$2
 MAX=$3
 SIZEI=`ls -l $FN | tr -s \  | cut -d\  -f5`
 if [ $((SIZEI)) -lt $((MIN)) ]
  then
   echo " ERROR: Image file: $FN: Smaller than expected. - Quiting.."
   exit 1
  fi
  if [ $((SIZEI)) -gt $((MAX)) ]
  then
   echo " ERROR: Image file: $FN: Larger than expected. - Quiting.."
   exit 1
  fi
}

check_bootloader_image() {
 if [ ! -f "$BIMG" ] ; then
  echo " ERROR: Image file: $BIMG: Not found. Quiting.."
  bi=0
  exit 1
 else 
  checkFile $BIMG
  check_file_size $BIMG 10000 $SIZEPB
 fi
}

# Verify image correctness
check_fpga_image() {
 if [ ! -f "$FIMG" ] ; then
  echo " ERROR: Image file: $FIMG: Not found. Quiting.."
  fi=0
  exit 1
 else 
  checkFile $FIMG
  ./mcstool $FIMG $TMPFBIN >/dev/null
  if [ $? -ne 0 ] ; then 
   echo " ERROR: Image file: $FIMG: Format error. Quiting.."
   fi=0
   exit 1
  fi
 fi
}

# Verify image correctness
check_kernel_image() {
 if [ ! -f "$KIMG" ] ; then
   ki=0
   if [ $ri -eq 1 ] ; then
     # allow bootloader or rootfs only update
     echo " WARNING:    Image: $KIMG: Not found. - Skipping.."
   else
     echo " ERROR: Image file: $KIMG: Not found. - Quiting.."
     exit 1
   fi
 else 
  checkFile $KIMG

  check_file_size $KIMG 4600000 $SIZEPK
 fi 
}

# Verify image correctness
check_root_image() {
 if [ ! -f "$RIMG" ] ; then
   ri=0
   if [ $ki -eq 1 -o $ai -eq 1 ] ; then
     # allow bootloader or kernel only update
     echo " WARNING:    Image: $RIMG: Not found. - Skipping.."
   else
     echo " ERROR: Image file: $RIMG: Not found. - Quiting.." 
     exit 1
   fi
 else 
  checkFile $RIMG
  # known sizes 040718 bs-root: 23672082, main-root: 124765893
  check_file_size $RIMG 23000000 $SIZEPR
 fi 
}

# Verify image correctness
check_all_image() {
 if [ ! -f "$AIMG" ] ; then
   ai=0
   if [ $ki -eq 1 -o $ri -eq 1 ] ; then
     # allow update with separate images
     echo " WARNING:    Image: $AIMG: Not found. - Skipping.."
   else
     echo " ERROR: Image file: $AIMG: Not found. - Quiting.." 
     exit 1
   fi
 else 
  # Other images cannot be updated at the same time with full image
  ki=0
  ri=0
  fi=0
  checkFile $AIMG
  check_file_size $AIMG $SIZEPR $(($SIZEPB + $SIZEPK*2 + $SIZEPRBS + $SIZEPR))
 fi 
}

# Verify files, iff
check_files() {
 get_sizes
 if [ $ci -eq 1 ] ; then
   check_addon_image
 fi

  if [ $fi -eq 1 ] ; then
    check_fpga_image
  fi
 if [ $bi -eq 1 ] ; then
    check_bootloader_image
  fi
  if [ $ki -eq 1 ] ; then
    check_kernel_image
  fi
  if [ $ri -eq 1 ] ; then
    check_root_image
  fi
#  if [ $ai -eq 1 ] ; then
#    check_all_image
#  fi
}

# Args: <filename> <source> <test-cmd>
copyhere ()
{
 if [ ! -x $1 ] ; then
  cp -f $2 .
 fi
 if [ ! -x $1 ] ; then
  echo "Error: Local version of '$1' missing. - Quiting.."
  exit 1
 fi
 ./$3 >/dev/null
 if [ $? -ne 0 ] ; then
  echo "Error: Nonworking version of local '$1'. - Quiting.."
  exit 2
 fi
}

copytmp ()
{
 if [ ! -x $1 ] ; then
  cp -f $2 /tmp
 fi
 #if [ ! -x $1 ] ; then
 # echo "Error: Local version of '$1' missing. - Quiting.."
 # exit 1
 #fi
 /tmp/$3 >/dev/null
 if [ $? -ne 0 ] ; then
  echo "Error: Nonworking version of local '$1'. - Quiting.."
  exit 2
 fi
}

# Init-like
prepare_environment() {
prepare_services
}

# Check for required tools
prepare_tools(){
 #copytmp busybox /bin/busybox "busybox echo 'good'"
 #if [ $fi -eq 1 ] ; then copytmp mcstool "/usr/sbin/mcstool" "mcstool" ; fi
    echo "Preparing tools" > /dev/null
}

# Check constrains for actions
check_limits(){
if [ `expr $ki + $ri + $ai + $ci + $fi` -eq 0 ]
then
 echo "ERROR: Cant flash without accepted set of images. - Quiting.."
 exit 1
fi
}

# Listing parts of actions to do 
show_effects() {
echo
echo  "       === Listing: Actions: ==="
if [ $fi -eq 1 ]
then
 decho 1 "          Updating: FPGA   "
 decho 2 "                    Dangerous if failed"
fi
if [ $bi -eq 1 ]
then
 decho 1 "          Updating: bootloader  "
fi
if [ $usrclear -eq 1 ]
then
 decho 1 "          Clearing: /usr/local  ALL FILES WILL BE LOST."
fi
if [ $ci -eq 1 ]
then
 decho 1 "           Install: $CIMG"
 if [ $usrclear -ne 1 ] ; then
   if [ $de -eq 1 ] ; then
     decho 1 "                    after clearing existing folders."
   else
     decho 1 "                    Overrides existing files."
   fi
 fi

fi
if [ $ki -eq 1 ]
then
 decho 1 "          Updating: kernel"
fi
if [ $ri -eq 1 ]
then
 decho 1   "          Updating: root-fs      ALL FILES ON ROOT-FS WILL BE LOST"
 if [ $usrclear -ne 1 ] ; then
   decho 2 "                    Files on /usr/local unaffected" 
 fi
fi
if [ $ai -eq 1 ]
then
 decho 1   "          Updating: full device OS  ALL FILES ON ROOT-FS WILL BE LOST"
fi
}

prepare_services(){
 decho 3 "         Preparing: just a minute.."
 #if [ -e /etc/init.d/logrotate ] ; then
 #/etc/init.d/logrotate stop 2>&1 >/dev/null; fi
 #/etc/rcS.d/S90powerd stop
 sleep 1
}

# Final Confirmation
confirm_user(){
 if [ $ask -eq 1 ] ; then
  echo -e "\n  === Verification: ARE YOU SURE ? ===\n"
  echo -e   "    - If yes, press Enter to continue."
  echo -e   "    - IF NOT, press CTRL+C now to cancel update."
  decho 2   "   This is you last chance to do cancel!"
  dechoe 1  "\n** IF YOU CONTINUE, DO NOT INTERRUPT THE PROCESS UNTIL READY **" 
  read a
 else   
  echo -e "\n      Verification: Skipped, continuing.."
 fi
 dechoe 2 "\n     === Now doing: Requested actions ==="

 dechoe 1 "\n * DO NOT BREAK PROCESS OR SHUTDOWN THE UNIT before update is complete. *\n"
 sleep 1
}

# Usr local clear
clear_usrlocal() {
 /bin/echo "Cleaning /usr/local .."
  rm -rf /usr/local/*
}

fix_rootfs() {
    # Try mounting /opt
    mount /dev/mmcblk0p4 /mnt

    if [ $? -eq 0 ] ; then
	umount /mnt # /opt created, now add it to mainsystem fstab

	mount /dev/mmcblk0p4 /mnt
	if [ $? -eq 0 ] ; then
	    echo "/dev/mmcblk0p4           /opt    ext4    defaults,noatime,data=ordered,barrier=1 0       0" >> /etc/fstab
	    umount /mnt
	else
	    echo "Failed to fix rootfs!"
	fi
    fi
}

##
## FLASH EM
##
flash_bootloader() {
/bin/echo "Copying new bootloader image.."
 dd if=$BIMG of=/dev/$BDEV seek=2 bs=512 && sync && sync
}

flash_kernel() {
/bin/echo "Copying new kernel image.."
mkdir -p /tmp/kernel
mount /dev/mmcblk0p1 /tmp/kernel
cp $KIMG /tmp/kernel/$KDEV
umount /tmp/kernel
}

flash_root() {
/bin/echo "Copying new root-fs image.."
mkfs.ext4 -F /dev/$RDEV
mount /dev/$RDEV /mnt
/bin/echo "Unpacking new root-fs image. This will take few minutes."
tar xvf $RIMG -C /mnt
umount /mnt
# For normal system rootfs, fstab needs to be modified to mount /opt since it's done dynamically
if [ "$RDEV" != "$RDEV2" ] ; then
    fix_rootfs
fi
}

flash_all() {
/bin/echo "Copying new full image.."
/bin/echo "Not supported atm."
#dd if=$AIMG of=/tmp/$BDEV bs=512 && sync && sync
}

do_reboot(){
# All done, rebooooot
/bin/echo "Completed. Rebooting.."
reboot
}

# YEY... main()

echo  "===== CrossControl: Device Update ===== "$V

set_defaults

read_arguments $@

check_environment 

prepare_tools

show_request

check_files

check_limits

prepare_environment

show_effects

confirm_user

if [ $usrclear -eq 1 ] ; then
 clear_usrlocal
fi

if [ $ci -eq 1 ] ; then
 addon_install
fi

if [ $fi -eq 1 ] ; then
 flash_fpga
 exit 1
fi

if [ $bi -eq 1 ] ; then
  flash_bootloader
fi

if [ $ki -eq 1 ] ; then
 flash_kernel
fi

if [ $ri -eq 1 ] ; then
 flash_root
fi

if [ $ai -eq 1 ] ; then
 flash_all
fi

if [ $noreboot -eq 0 ] ; then
 do_reboot
fi

/bin/echo -e "\nEnd of all actions, H o l d    o n    t i g h t \n"
